#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>

// #define MIN3(a, b, c) ((a) < (b) ? ((a) < (c) ? (a) : (c)) : ((b) < (c) ? (b) : (c)))
#define DEL_COST (1)
#define INS_COST (1)
#define SUBS_COST (1)
#define KEEP_COST (0)
typedef enum {NONE, DEL, INS, SUBS} OpsT;

/** \brief Allocate a 2d unsigned int matrix
 *
 * \param rows number of rows
 * \param cols number of cols
 * \return pointer allocated matrix if successful; if not it aborts execution
 *
 */
unsigned int **allocUIntMatrix(size_t rows, size_t cols)
{
    unsigned int **mat = (unsigned int **)malloc(rows * sizeof(unsigned int *));
    if (!mat)
    {
        printf("%s: %s", __FILE__, __LINE__);
        abort();
    }
    mat[0] = (unsigned int *) malloc(rows * cols * sizeof(unsigned int));
    if (!mat[0])
    {
        printf("%s: %s", __FILE__, __LINE__);
        abort();
    }
    for (unsigned int i = 1; i < rows; i++)
    {
        mat[i] = mat[0] + i * cols;
    }
    return mat;
}

/** \brief Calculate minimum cost of three ops and set performed op
 *
 * \param del min for DEL
 * \param ins min for INS
 * \param subs min for SUBS
 * \param op operation code for op performed
 * \return min value
 *
 */
unsigned int min3(unsigned int del, unsigned int ins, unsigned int subs, OpsT *op)
{
    int min3 = INT_MAX;
    if (subs < del && subs < ins)
    {
        min3 = subs;
        *op = SUBS;
    }
    else if (ins < del)
    {
        min3 = ins;
        *op = INS;
    }
    else
    {
        min3 = del;
        *op = DEL;
    }
    return min3;
}

/** \brief Print cost matrix
 *
 * \param s1 first string
 * \param s2 2nd string
 * \param matrix cost matrix
 * \return void
 *
 */
void printCostMatrix(char *s1, char *s2, unsigned int **matrix)
{
    size_t len1 = strlen(s1);
    size_t len2 = strlen(s2);
    // print top heading
    printf("\n    ");
    for (size_t i=1; i <= len2; i++)
        printf(" %c", s2[i-1]);
    printf("\n  ");
    // print line zero
    for (size_t j=0; j <= len2; j++)
        printf("%2d", matrix[0][j]);
    puts("");
    // print rest
    for (size_t i=1; i <= len1; i++)
    {
        printf("%c ", s1[i-1]);
        for (size_t j=0; j <= len2; j++)
            printf("%2d", matrix[i][j]);
        puts("");
    }
}
// translate opCode to char for printing
static char translate(unsigned int val)
{
    char opCode;
    switch (val)
    {
    case INS:
        opCode = 'I';
        break;
    case DEL:
        opCode = 'D';
        break;
    case SUBS:
        opCode = 'S';
        break;
    case NONE:
        opCode = '+';
        break;
    default:
        opCode = '*';
    }
    return opCode;
}
/** \brief Print operations matrix
 *
 * \param s1 first string
 * \param s2 2nd string
 * \param p operations matrix (ops are: NONE, INS, DEL, SUBS)
 * \return void
 *
 */
void printOpsMatrix(char *s1, char *s2, OpsT **p)
{
    size_t len1 = strlen(s1);
    size_t len2 = strlen(s2);
    // print top heading
    printf("\n    ");
    for (size_t i=0; i <= len2; i++)
        printf(" %c", s2[i]);
    printf("\n  ");
    // print line zero
    for (size_t j=0; j <= len2; j++)
    {
        printf(" %c", translate(p[0][j]));
    }
    puts("");
    // print rest
    for (size_t i=1; i <= len1; i++)
    {
        printf("%c ", s1[i-1]);
        for (size_t j=0; j <= len2; j++)
            printf(" %c", translate(p[i][j]));
        puts("");
    }
}
/** \brief Retrieves the ops performed to get min cost edit
 *
 * \param s1 first string
 * \param s2 2nd string
 * \param p operations matrix (ops are: NONE, INS, DEL, SUBS)
 * \param i current row
 * \param j current column
 * \return void
 *
 */
void recPath(char *s1, char *s2, OpsT **p, size_t i, size_t j)
{
    if (i > 0 || j > 0)
    {
        switch(p[i][j])
        {
        case DEL:
            recPath(s1, s2, p, i-1, j);
            printf("D: %c -- _\n", (i == 0)? '_': s1[i-1]);
            break;
        case INS:
            recPath(s1, s2, p, i, j-1);
            printf("I: _ -- %c\n", (j == 0)? '_': s2[j-1]);
            break;
        case SUBS:
            recPath(s1, s2, p, i-1, j-1);
            printf("S: %c -- %c\n", (i == 0)? '_': s1[i-1], (j == 0)? '_': s2[j-1]);
            break;
        case NONE:
            recPath(s1, s2, p, i-1, j-1);
            printf("+: %c -- %c\n", (i == 0)? '_': s1[i-1], (j == 0)? '_': s2[j-1]);
            break;
        default:
            printf("ERROR: %c -- %c\n", (i == 0)? '_': s1[i-1], (j == 0)? '_': s2[j-1]);
            break;
        }
    }
}
/** \brief Calculates edit distance (Levenshtein) matrix
 *
 * \param s1 first char string
 * \param s2 3nd char string
 * \param p  matrix of operations performed to get min cost
 * \return matrix of min cost
 *
 */
unsigned int **levenshtein(char *s1, char *s2, OpsT ***p)
{
    size_t len1 = strlen(s1);
    size_t len2 = strlen(s2);
    unsigned int **d = allocUIntMatrix(len1 + 1, len2 + 1);
    *p = (OpsT **)allocUIntMatrix(len1 + 1, len2 + 1);

    d[0][0] = 0;
    for(unsigned int i = 1; i <= len1; ++i)
        d[i][0] = i;
    for(unsigned int i = 1; i <= len2; ++i)
        d[0][i] = i;

    for(unsigned int i = 1; i <= len1; ++i)
    {
        for(unsigned int j = 1; j <= len2; ++j)
        {
            if (s1[i-1] == s2[j-1])
            {
                d[i][j] = d[i - 1][j - 1] + KEEP_COST;
                (*p)[i][j] = NONE;
            }
            else
            {
                OpsT op;
                d[i][j] = min3( d[i - 1][j] + DEL_COST, d[i][j - 1] + INS_COST,
                                d[i - 1][j - 1] + SUBS_COST, &op );
                (*p)[i][j] = op;
            }
        }
    }
    // return d[len1][len2];
    return d;
}

int main()
{
    char *s1 = "algorithm";
    char *s2 = "altruistic";
    unsigned int **p;
    unsigned int **E;
    char s1read[1024];
    char s2read[1024];

    printf("s1=");
    scanf("%1024s", s1read);
    // if s1read too short then go on default values
    if (strlen(s1read) > 1)
    {
        s1 = s1read;
        printf("s2=");
        scanf("%1024s", s2read);
        s2 = s2read;
    }
    size_t len1 = strlen(s1);
    size_t len2 = strlen(s2);
    E = levenshtein(s1, s2, &p);
    printf("E[m, n]=%d\n", E[len1][len2]);
    printCostMatrix(s1, s2, E);
    printOpsMatrix(s1, s2, p);
    recPath(s1, s2, p, len1, len2);
    return 0;
}
